#include <stdlib.h>
#include <Wire.h>
#include <EEPROM.h>

#include "bmp280.h"
#include "SparkFun_SCD30_Arduino_Library.h"
#include "SparkFunCCS811.h"

#define PIN_MQ2		A1
#define PIN_DUST	A0
#define PIN_TRIG	13
#define PIN_MQ2PWM	10
#define PIN_FANWR	6
#define PIN_FANRD	9

#define SECONDIV		1000
#define FANPWM			64

#define CCS811_ADDR		0x5B

#define CALIB_DUST_OFFSET	1300
#define CALIB_DUST_FACTOR	0.000117750901

//             8191
// Rx = RL * (------ - 1)
//             adc
//
// Disregarding the 9.8 factor (in real R calculation)
// Rs/R0 == 1.00 at air
// Rs/R0 ~= 0.08 at 10000 ppm
//
// gas = 10^(-Rs/R0 + Fconst)

#define CALIB_MQ2_0		950
#define CALIB_FCONST		4.3655261495981
#define RS(v)			((8191.0/(v)) - 1.0)
#define R0			RS(calib.val.mq2_r0)

typedef struct __calibvals_t {
	struct {
		uint16_t dust_offset;
		uint16_t mq2_r0;
		float dust_factor;
		float mq2_factor;
		uint16_t ccs_baseline;
		uint8_t fan;
		uint8_t reserved;
	} val;
	uint32_t crc;
} calibvals_t;

static uint16_t dust;
static uint16_t mq2;
static uint32_t eqco2;
static uint32_t eqvoc;
static calibvals_t calib;

static Adafruit_BMP280 bmp280;
static SCD30 scd30;
static CCS811 ccs811(CCS811_ADDR);

static float r0;
static float pdsum;
static int oldpd;
static int nsum;

static uint8_t fanstatus;
static uint16_t fantag;
static uint8_t fanspeed;

static char sbuf[31+1];
static uint8_t nsbuf;
static bool neednl;
static uint8_t initwait;

static uint32_t mem_crc(const void *p, uint16_t len)
{
	static const uint32_t crc_table[16] PROGMEM = {
		0x00000000, 0x1db71064, 0x3b6e20c8, 0x26d930ac,
		0x76dc4190, 0x6b6b51f4, 0x4db26158, 0x5005713c,
		0xedb88320, 0xf00f9344, 0xd6d6a3e8, 0xcb61b38c,
		0x9b64c2b0, 0x86d3d2d4, 0xa00ae278, 0xbdbdf21c
	};

	uint32_t crc = ~0L;

	for(uint16_t index = 0 ; index < len; ++index) {
		uint8_t val = ((const uint8_t *)p)[index];
		crc = pgm_read_dword(&crc_table[(crc ^ val) & 0x0f]) ^ (crc >> 4);
		crc = pgm_read_dword(&crc_table[(crc ^ (val >> 4)) & 0x0f]) ^ (crc >> 4);
		crc = ~crc;
	}
	return crc;
}

static void saveCalib()
{
	calib.crc = mem_crc(&calib.val, sizeof(calib.val));
	EEPROM.put(0, calib);
}

//
// BL<int>	- set CCS811 baseline
// Bl<int>	- set CCS811 baseline (no save)
// C		- show values
// c		- show values
// DO<int>	- set dust offset
// Do<int>	- set dust offset (no save)
// DF<flt>	- set dust factor
// Df<flt>	- set dust factor (no save)
// GO<int>	- set gas offset
// Go<int>	- set gas offset (no save)
// GF<flt>	- set gas factor
// Gf<flt>	- set gas factor (no save)
// FS<int>	- set fan speed
// Fs<int>	- set fan speed (no save)
//
static void handleSerial()
{
	bool save = false;
	char *eptr;
	switch(sbuf[0]) {
	case 'B':
		if(sbuf[1] == 'L' || sbuf[1] == 'l') {
			uint16_t i = strtoul(&sbuf[2], &eptr, 0);
			if(!*eptr) {
				calib.val.ccs_baseline = i;
				save = sbuf[1] == 'L';
			}
		}
		break;

	case 'c':
	case 'C':
		Serial.write('C');
		Serial.write(' ');
		Serial.print(calib.val.dust_offset);
		Serial.write(' ');
		Serial.print(calib.val.dust_factor, 12);
		Serial.write(' ');
		Serial.print(calib.val.mq2_r0);
		Serial.write(' ');
		Serial.print(calib.val.mq2_factor, 12);
		Serial.write(' ');
		Serial.print(calib.val.ccs_baseline);
		Serial.write(' ');
		Serial.print(ccs811.getBaseline());
		Serial.write(' ');
		Serial.print(calib.val.fan);
		Serial.println();
		break;

	case 'D':
		if(sbuf[1] == 'O' || sbuf[1] == 'o') {
			uint16_t i = strtoul(&sbuf[2], &eptr, 0);
			if(!*eptr) {
				calib.val.dust_offset = i;
				save = sbuf[1] == 'O';
			}
		} else if(sbuf[1] == 'F' || sbuf[1] == 'f') {
			float f = strtod(&sbuf[2], &eptr);
			if(!*eptr) {
				calib.val.dust_factor = f;
				save = sbuf[1] == 'F';
			}
		}
		break;

	case 'F':
		if(sbuf[1] == 'S' || sbuf[1] == 's') {
			uint16_t i = strtoul(&sbuf[2], &eptr, 0);
			if(!*eptr && i <= 255) {
				calib.val.fan = i;
				analogWrite(PIN_FANWR, fanspeed = calib.val.fan);
				save = sbuf[1] == 'S';
			}
		}
		break;

	case 'G':
		if(sbuf[1] == 'O' || sbuf[1] == 'o') {
			uint16_t i = strtoul(&sbuf[2], &eptr, 0);
			if(!*eptr) {
				calib.val.mq2_r0 = i;
				r0 = R0;
				save = sbuf[1] == 'O';
			}
		} else if(sbuf[1] == 'F' || sbuf[1] == 'f') {
			float f = strtod(&sbuf[2], &eptr);
			if(!*eptr) {
				calib.val.mq2_factor = f;
				save = sbuf[1] == 'F';
			}
		}
		break;
	}

	if(save) {
		saveCalib();
	}
}

void setup()
{
	pinMode(PIN_TRIG, OUTPUT);

	pinMode(PIN_MQ2PWM, OUTPUT);
	analogWrite(PIN_MQ2PWM, 255);

	pinMode(PIN_FANRD, INPUT_PULLUP);
	fanstatus = digitalRead(PIN_FANRD);
	fantag = millis();
	pinMode(PIN_FANWR, OUTPUT);

	Serial.begin(38400);

	EEPROM.get(0, calib);
	uint32_t crc = mem_crc(&calib.val, sizeof(calib.val));
	if(crc != calib.crc) {
		calib.val.dust_offset = CALIB_DUST_OFFSET;
		calib.val.dust_factor = CALIB_DUST_FACTOR;
		calib.val.mq2_r0 = CALIB_MQ2_0;
		calib.val.mq2_factor = CALIB_FCONST;
		calib.val.ccs_baseline = 0;
		calib.val.fan = FANPWM;
		saveCalib();
	}

	analogWrite(PIN_FANWR, fanspeed = calib.val.fan);

	// XXX
	// Adafruit original has a pull-up on SDO
	//bmp280.begin(BMP280_ADDRESS);
	// The Arduinotech clone has a pull-down on SDO
	bmp280.begin(BMP280_ADDRESS_ALT);

	scd30.begin();
	scd30.softReset();
	scd30.setTemperatureOffset(0.0);
	scd30.setAutoSelfCalibration(false);
	scd30.setMeasurementInterval(2);

	ccs811.begin();
	if(calib.val.ccs_baseline)
		ccs811.setBaseline(calib.val.ccs_baseline);
	r0 = R0;

	pdsum = 0.0;
	oldpd = 1013;
	nsum = 0;
	nsbuf = 0;
	neednl = true;

	initwait = 15;
}

void loop()
{
	static uint32_t seconds;
	static uint16_t prevtime;
	uint16_t now = millis();
	uint16_t tdiff = now - prevtime;
	if(tdiff >= SECONDIV) {
		prevtime += SECONDIV;
		// Start with timing sensitive part. The loop starts here when
		// the timer interrupt has fired and is not expected for
		// another millisecond. Therefore, the timing should be stable.
		digitalWrite(PIN_TRIG, 1);
		delayMicroseconds(280);
		dust = (((dust << 3) - dust) >> 3) + analogRead(PIN_DUST);
		delayMicroseconds(40);
		digitalWrite(PIN_TRIG, 0);

		seconds++;
		if(seconds >= 7 * 86400) {
			uint16_t bl = ccs811.getBaseline();
			if(bl) {
				calib.val.ccs_baseline = bl;
				saveCalib();
			}
			seconds = 0;
		}

		mq2 = (((mq2 << 3) - mq2) >> 3) + analogRead(PIN_MQ2);

		float p = bmp280.readPressure();
		float t = bmp280.readTemperature();
		static float co2, tmp, hum;

		if(scd30.dataAvailable()) {
			co2 = scd30.getCO2();
			tmp = scd30.getTemperature();
			hum = scd30.getHumidity();
		}

		float gas = 0.0;
		if(mq2 != 0) {
			gas = pow(10.0, (-RS(mq2)/r0 + 1.0)*calib.val.mq2_factor);
		}

		float dst = 0.0;
		if(dust >= calib.val.dust_offset) {
			dst = (dust - calib.val.dust_offset) * calib.val.dust_factor;
		}

		if(ccs811.dataAvailable()) {
			ccs811.readAlgorithmResults();
			eqco2 = (((eqco2 << 3) - eqco2) >> 3) + ccs811.getCO2();
			eqvoc = (((eqvoc << 3) - eqvoc) >> 3) + ccs811.getTVOC();
		}

		pdsum += p * 0.01;	// Needs to be in mBar
		nsum++;
		if(nsum >= 60) {
			pdsum /= nsum;
			int newpd = (int)round(pdsum);
			if(abs(oldpd - newpd) >= 2) {
				scd30.setAmbientPressure(newpd);
				oldpd = newpd;
			}
			nsum = 0;
			pdsum = 0.0;
		}

		if(!initwait) {
			Serial.write('D');
			Serial.write(' ');
			Serial.print(co2, 3);
			Serial.write(' ');
			Serial.print(tmp, 3);
			Serial.write(' ');
			Serial.print(hum, 3);
			Serial.write(' ');
			Serial.print(p, 3);
			Serial.write(' ');
			Serial.print(t, 3);
			Serial.write(' ');
			Serial.print(gas, 3);
			Serial.write(' ');
			Serial.print(dst, 6);
			Serial.write(' ');
			Serial.print(mq2);
			Serial.write(' ');
			Serial.print(dust);
			Serial.write(' ');
			Serial.print((float)eqco2 * 0.125, 3);
			Serial.write(' ');
			Serial.print((float)eqvoc * 0.125, 3);
			Serial.write(' ');
			Serial.print(fanspeed);
			Serial.println();
		} else {
			initwait--;
		}
	}

	if(Serial.available() > 0) {
		char c = Serial.read();
		if(c == '\n') {
			if(!neednl)
				handleSerial();
			memset(sbuf, 0, sizeof(sbuf));
			neednl = false;
			nsbuf = 0;
		} else if(isalnum(c) || c == '.' || c == ' ') {
			if(nsbuf < sizeof(sbuf)-1) {
				sbuf[nsbuf] = c;
				nsbuf++;
				sbuf[nsbuf] = '\0';
			} else {
				memset(sbuf, 0, sizeof(sbuf));
				neednl = true;
				nsbuf = 0;
			}
		} else if(c != '\r') {
			memset(sbuf, 0, sizeof(sbuf));
			neednl = true;
			nsbuf = 0;
		}
	}
}
